* * * * * *    Methods of QTL linkage analysis of sib-pairs       * * * * * *
-----------------------------------------------------------------------------
Method  Descriptions
-----------------------------------------------------------------------------
 1*     mixture
 2*     pi-hat
 3*     mixture,                       adjustment by conditioning on trait
 4*     pi-hat,                        adjustment by conditioning on trait
 5.     mixture, without imputation
 6.     pi-hat, without imputation
 7.     mixture, with imputation
 8.     pi-hat, with imputation
 9.     mixture, without imputation, adjustment by integral
10.     pi-hat, without imputation, adjustment by integral
11.     mixture, without imputation, adjustment by approximation to integral
12.     pi-hat, without imputation, adjustment by approximation to integral
13.     mixture, without imputation, adjustment by conditioning on trait
14.     pi-hat, without imputation, adjustment by conditioning on trait
-----------------------------------------------------------------------------
legend:  * Complete data, . Selected data.
;

/*****************************************************/
/* Parameters used in the program                    */
/*---------------------------------------------------*/
/* t1,t2 -- the cut points, user-specified           */
/* v     -- the trait variance, user-specified       */
/* c     -- the sib covariance, user-specified       */
/* y     -- the QTL variance, to be estimated        */
/* - v,c are now calculated according to raw data  - */
/*   and the traits are standardized                 */
/*---------------------------------------------------*/
/* This program uses complete dataset which contains */
/* x1,x2: trait values of sibs 1 and 2               */
/* x3,x4: proportions of sharing 0 and 1 alleles IBD */
/* then selects sib-pairs according to cut-offs      */
/*****************************************************/

%let  t1=-0.3;
%let  t2=0.3;
%let  v=1;
%let  y=0.1;

%macro step;

data gemini;
  set x.traitibd (rename=(trait1=x1 trait2=x2 z0=x3 z1=x4 z2=x5));
  where marker=&marker;
  x1x2=x1+x2;
  if x1x2=. then delete;
  drop x1x2;
  pihat=1-(x3+x4/2);
run;

proc corr noprint cov outp=simple;
run;

data all;
  set gemini;
run;

proc standard data=all out=abc mean=0 std=1;
  var x1 x2;
run;

data abc;
  set abc;
  t1=&t1;
  t2=&t2;
  select=((x1>t2 & x2>t2)|(x1<t1 & x2<t1)|(x1>t2 & x2<t1)|(x1<t1 & x2>t2));
  keep x1-x4 pihat select;
run;
proc means data=abc;
  where select=1;
  var pihat;
  output out=ssimple mean=pihatbar std=pihatstd;
run;
data check;
  set abc;
  y=x1*x2;
run;
proc means mean std;
  var y x1 x2 pihat;
run;
proc datasets lib=work;
  delete all check;
quit;

proc iml symsize=300000;

START dnorm(y) global(x,pi,pihatbar,v,c);
  sigma_q=y;
  sigma11=v;
  sigma22=v;
  sigma12=c+(pi-0.5)*sigma_q**2;
  sigma=(sigma11||sigma12)//(sigma12||sigma22);
  sigma_d=det(sigma);
  sigma_i=inv(sigma);
  l=sqrt(1/sigma_d)*exp(-x*sigma_i*x`/2);
  return (l);
FINISH dnorm;

START objf1(y) global(data,pi,x);
  n=nrow(data);
  l=0;
  do i=1 to n;
     x=(data[i,1]||data[i,2]);
     z0=data[i,3];
     z1=data[i,4];
     z2=1-z0-z1;
     pi=0;   l0=dnorm(y);
     pi=0.5; l1=dnorm(y);
     pi=1;   l2=dnorm(y);
     l=l+log(z0*l0+z1*l1+z2*l2);
  end;
  return (-l);
FINISH objf1;

START objf2(y) global(data,pi,x);
  n=nrow(data);
  l=0;
  do i=1 to n;
     x=(data[i,1]||data[i,2]);
     z0=data[i,3];
     z1=data[i,4];
     z2=1-z0-z1;
     pi=z1/2+z2; lhat=dnorm(y);
     l=l+log(lhat);
  end;
  return (-l);
FINISH objf2;

START objf3(y) global(data,pi,x);
  n=nrow(data);
  l=0;
  a=0;
  do i=1 to n;
     x=(data[i,1]||data[i,2]);
     z0=data[i,3];
     z1=data[i,4];
     z2=1-z0-z1;
     pi=0;   l0=dnorm(y);
     pi=0.5; l1=dnorm(y);
     pi=1;   l2=dnorm(y);
     a=a+log(0.25*l0+0.5*l1+0.25*l2);
     l=l+log(z0*l0+z1*l1+z2*l2);
  end;
  return (-(l-a));
FINISH objf3;

START objf4(y) global(data,pi,x);
  n=nrow(data);
  l=0;
  a=0;
  do i=1 to n;
     x=(data[i,1]||data[i,2]);
     z0=data[i,3];
     z1=data[i,4];
     z2=1-z0-z1;
     pi=z1/2+z2;  lhat=dnorm(y);
     l=l+log(lhat);
     pi=0;   l0=dnorm(y);
     pi=0.5; l1=dnorm(y);
     pi=1;   l2=dnorm(y);
     a=a+log(0.25*l0+0.5*l1+0.25*l2);
  end;
  return (-(l-a));
FINISH objf4;

START objf5(y) global(seldata,pi,x);
  n=nrow(seldata);
  l=0;
  do i=1 to n;
     x=(seldata[i,1]||seldata[i,2]);
     z0=seldata[i,3];
     z1=seldata[i,4];
     z2=1-z0-z1;
     pi=0;   l0=dnorm(y);
     pi=0.5; l1=dnorm(y);
     pi=1;   l2=dnorm(y);
     l=l+log(z0*l0+z1*l1+z2*l2);
  end;
  return (-l);
FINISH objf5;

START objf6(y) global(seldata,pi,x);
  n=nrow(seldata);
  l=0;
  do i=1 to n;
     x=(seldata[i,1]||seldata[i,2]);
     z0=seldata[i,3];
     z1=seldata[i,4];
     z2=1-z0-z1;
     pi=z1/2+z2;    lhat=dnorm(y);
     l=l+log(lhat);
  end;
  return (-l);
FINISH objf6;

START objf7(y) global(data,pi,x);
  n=nrow(data);
  l=0;
  do i=1 to n;
     x=(data[i,1]||data[i,2]);
     sel=data[i,5];
     if sel=1 then do;
        z0=data[i,3];
        z1=data[i,4];
        z2=1-z0-z1;
     end;
     else do;
        z0=0.25;
        z1=0.5;
        z2=0.25;
     end;
     pi=0;   l0=dnorm(y);
     pi=0.5; l1=dnorm(y);
     pi=1;   l2=dnorm(y);
     l=l+log(z0*l0+z1*l1+z2*l2);
  end;
  return (-l);
FINISH objf7;

START objf8(y) global(data,pi,x);
  n=nrow(data);
  l=0;
  do i=1 to n;
     x=(data[i,1]||data[i,2]);
     sel=data[i,5];
     if sel=1 then do;
        z0=data[i,3];
        z1=data[i,4];
        z2=1-z0-z1;
     end;
     else do;
        z0=0.25;
        z1=0.5;
        z2=0.25;
     end;
     pi=z1/2+z2;  lhat=dnorm(y);
     l=l+log(lhat);
  end;
  return (-l);
FINISH objf8;

START bnorm(y) global(pi,pihatbar,v,c,t1,t2);
  sigma_q=y;
  sigma11=v;
  sigma22=v;
  sigma12=c+(pi-0.5)*sigma_q**2;
  sigma21=c+(pi-0.5)*sigma_q**2;
  d1=t1/sqrt(v);
  d2=t2/sqrt(v);
  rho=sigma12/v;
  s1=probbnrm(d1,d1,rho);
  s2=1-2*cdf('norm',d2)+probbnrm(d2,d2,rho);
  s3=cdf('norm',d1)-probbnrm(d1,d2,rho);
  return (s1+s2+2*s3);
FINISH bnorm;

START objfc(y) global(seldata,pi,x,v,c,method);
  n=nrow(seldata);
  l=0;
  a=0;
  do i=1 to n;
     z0=seldata[i,3];
     z1=seldata[i,4];
     z2=1-z0-z1;
     pi=z1/2+z2;
     x=(seldata[i,1]||seldata[i,2]);
     if (method=10|method=12|method=14) then l=l+log(dnorm(y));
     pi=0;   l0=dnorm(y);
     pi=0.5; l1=dnorm(y);
     pi=1;   l2=dnorm(y);
     if (method=9|method=11|method=13) then l=l+log(z0*l0+z1*l1+z2*l2);
     w=(0.25*l0+0.5*l1+0.25*l2);
     if (method=11|method=12) then a=a+w/l1;
     if (method=13|method=14) then a=a+log(w);
  end;

  if (method=9|method=10) then do;
     pi=0;   a0=bnorm(y);
     pi=0.5; a5=bnorm(y);
     pi=1;   a1=bnorm(y);
     a=(0.25*a0+0.5*a5+0.25*a1);
  end;

  if (method=9|method=10) then a=n*log(a);
  if (method=11|method=12) then a=n*log(a);
  if (method=13|method=14) then a=a;

  return(-(l-a));
FINISH objfc;

start regress(x,y,name) global(b,dfe,xpxi,q,mse,x2b,f,lrt);
  n=nrow(x);                  /* number of observations   */
  k=ncol(x);                  /* number of variables      */
  xpx=x`*x;                   /* cross-products           */
  xpy=x`*y;
  xpxi=inv(xpx);              /* inverse crossproducts    */
  b=xpxi*xpy;                 /* parameter estimates      */

  yhat=x*b;                   /* predicted values         */
  resid=y-yhat;               /* residuals                */
  ssy=y`*y;                   /* total sum of squares     */
  sse=resid`*resid;           /* sum of squared errors    */
  dfe=n-k;                    /* degrees of freedom error */
  mse=sse/dfe;                /* mean squared error       */
  rmse=sqrt(mse);             /* root mean squared error  */

  lrt=n*(log(ssy)-log(sse));  /* likelihood ratio chi-sq  */
  lambda=exp(-lrt/2);         /* likelihood ratio         */
  f=(n-2)*(lambda**(-2/n)-1); /* F test                   */
  if b[1]<0 then lrt=0;       /* only for reg thru origin */
  if b[1]<0 then f=0;
  covb=xpxi#mse;              /* covariance of estimates  */
  stdb=sqrt(vecdiag(covb));   /* standard errors          */
  t=b/stdb;                   /* ttest for estimates=0    */
  x2b=t#t;
  probt=1-probf(x2b,1,dfe);   /* significance probability */
  do i=1 to k;                /* 1-tail chi<?>-square     */
    if (t[i]<0) then x2b[i]=0;
  end;
* print name b stdb t probt;
  s=diag(1/stdb);
  corrb=s*covb*s;             /* correlation of estimates */
* print ,"covariance of estimates",  covb[r=name c=name],
         "correlation of estimates",corrb[r=name c=name];

  if nrow(tval)=0 then return;/* is a t-value specified?  */
  projx=x*xpxi*x`;            /* hat matrix               */
  vresid=(i(n)-projx)*mse;    /* covariance of residuals  */
  vpred =projx#mse;           /* covariance of predicted values */
  h=vecdiag(projx);           /* hat leverage values      */
  lowerm=yhat-tval#sqrt(h*mse); /*lower confidence limit for mean*/
  upperm=yhat+tval#sqrt(h*mse);     /* upper */
  lower =yhat-tval#sqrt(h*mse+mse);/*lower confidence limit for ind*/
  upper =yhat+tval#sqrt(h*mse+mse); /* upper */
  print ,,"predicted values, residuals, and limits" ,,
         y yhat resid h lowerm upperm lower upper;
finish;

start ltest;
  dfn=nrow(l);
  lb=l*b;
  vlb=l*xpxi*l`;
  q=lb`*inv(vlb)*lb /dfn;
  f=q/mse;
  prob=1-probf(f,dfn,dfe);
  print ,f dfn dfe prob;
finish;

use abc;
read all var {x1 x2 x3 x4 select} into data;
read all where (select=1) var {x1 x2 x3 x4 select} into seldata;
close abc;

use simple;
read all where (_type_='CORR' & _name_='X1') var{x2} into c;
read all where (_type_='MEAN') var{x1 x2 pihat} into mean;
read all where (_type_='STD') var{x1 x2 pihat} into std;
close simple;
%*summary var{x1 x2 x3 x4 x1x2}  stat{uss mean std} where{select=1} opt{save};
%*need calculation for c based on x1x2;

use ssimple;
read all var{pihatbar pihatstd} into smstd;
close ssimple;
print 'the correlation, mean & std ' c mean std smstd;
%*summary var{pihat} stat{mean std} where{select=1} opt{save};

t1=&t1;
t2=&t2;
v=&v;
y=&y;

opt={0 0};
con={0 . ., 0.95 . .};

r2=j(1,4,0);chi=j(1,4,0);sf=j(1,4,0);

/*
r2=j(1,18,0);chi=j(1,22,0);sf=j(1,14,0);
*/

%do j=1 %to 4;
  method=&j;
  %NLPCG;
%end;

/*
n=nrow(data);
v=j(n,1,0);
u=j(n,2,0);
name={"intercept","pihat"};
do i=1 to n;
  v[i]=data[i,1]*data[i,2];
  u[i,1]=1;
  u[i,2]=1-(data[i,3]+data[i,4]/2);
end;
free data;
run regress(u,v,name);
r2[15]=b[2];
chi[15]=x2b[2];

n=nrow(seldata);
v=j(n,1,0);
u=j(n,2,0);
v1=j(n,1,0);
u1=j(n,1,0);
do i=1 to n;
  v[i]=seldata[i,1]*seldata[i,2];
  u[i,1]=1;
  u[i,2]=1-(seldata[i,3]+seldata[i,4]/2);
  v1[i]=v[i]-c;
  u1[i]=u[i,2]-0.5;
end;
free seldata;
run regress(u,v,name);
r2[16]=b[2];
chi[16]=x2b[2];
run regress(u1,v1,name);
r2[17]=b[1];
chi[17]=lrt;
chi[18]=f;
chi[19]=x2b[1];
run regress(v1,u1,name);
r2[18]=b[1]*(1+c**2)/(1/8);
chi[20]=lrt;
chi[21]=f;
chi[22]=x2b[1];

reset linesize=132 pagesize=5000;
mattrib r2 label={'qtl variance'} chi label={'lrt chi^2 statistics'};
mattrib r2 label={''} chi label={''} ;
print 'loop' &i, r2 [format=7.3], chi [format=7.1];

rname='v1':'v18';
chiname='chi1':'chi22';
sfname='sf1':'sf14';
rchiname=rname||chiname||sfname;
rchisf=r2||chi||sf;
create step from rchisf [colname=rchiname];
append from rchisf;
*/

reset linesize=132 pagesize=5000;
mattrib r2 label={'qtl variance'} chi label={'lrt chi^2 statistics'};
mattrib r2 label={''} chi label={''} ;
print 'loop' &i, r2 [format=7.3], chi [format=7.1];

rname='v1':'v4';
chiname='chi1':'chi4';
sfname='sf1':'sf4';
rchiname=rname||chiname||sfname;
rchisf=r2||chi||sf;
create step from rchisf [colname=rchiname];
append from rchisf;

quit;

%mend step;

%macro NLPCG;
  if &j<5 then pihatbar=mean[3];
  else pihatbar=smstd[1];
  if &j<9 then do;
    call NLPCG(rc,r,"objf&j",y,opt,con);
    chi[&j]=-2*(objf&j(r)-objf&j(0));
  end;
  else do;
    call NLPCG(rc,r,"objfc",y,opt,con);
    chi[&j]=-2*(objfc(r)-objfc(0));
  end;
  r2[&j]=r**2;
  sf[&j]=rc;
%mend;

%macro loop;
  %do i=1 %to 31;    %let seed=&seed0+&i*100;
     %if &i=1 %then %let        marker=  0.0;
     %else %if &i= 2 %then %let marker=  0.6;
     %else %if &i= 3 %then %let marker=  6.6;
     %else %if &i= 4 %then %let marker= 14.8;
     %else %if &i= 5 %then %let marker= 22.0;
     %else %if &i= 6 %then %let marker= 23.2;
     %else %if &i= 7 %then %let marker= 32.8;
     %else %if &i= 8 %then %let marker= 36.4;
     %else %if &i= 9 %then %let marker= 37.3;
     %else %if &i=10 %then %let marker= 44.0;
     %else %if &i=11 %then %let marker= 44.1;
     %else %if &i=12 %then %let marker= 44.2;
     %else %if &i=13 %then %let marker= 49.5;
     %else %if &i=14 %then %let marker= 49.6;
     %else %if &i=15 %then %let marker= 50.3;
     %else %if &i=16 %then %let marker= 52.0;
     %else %if &i=17 %then %let marker= 54.0;
     %else %if &i=18 %then %let marker= 54.6;
     %else %if &i=19 %then %let marker= 55.1;
     %else %if &i=20 %then %let marker= 61.6;
     %else %if &i=21 %then %let marker= 63.8;
     %else %if &i=22 %then %let marker= 65.1;
     %else %if &i=23 %then %let marker= 65.2;
     %else %if &i=24 %then %let marker= 67.7;
     %else %if &i=25 %then %let marker= 75.8;
     %else %if &i=26 %then %let marker= 84.3;
     %else %if &i=27 %then %let marker= 95.0;
     %else %if &i=28 %then %let marker= 99.4;
     %else %if &i=29 %then %let marker=104.8;
     %else %if &i=30 %then %let marker=117.8;
     %else %if &i=31 %then %let marker=128.8;
     %step;
     proc append base=savestep data=step;
     run;
  %end;
  data x.opt;
    set savestep;
    if _n_=1 then delete;
  run;
  title 'The result after all the loops';
  proc print data=x.opt round;
  run;
%mend loop;

data savestep;
  v1=.;chi1=.;sf1=.;
  v2=.;chi2=.;sf2=.;
  v3=.;chi3=.;sf3=.;
  v4=.;chi4=.;sf4=.;
run;

options nocenter nodate nonumber mprint;

%let seed0=2000;

libname x 'c:\iop\sas';

title;
%loop;
